#
#  opt_func.tcl -- Generic optimization procedure functions.
#
#  Written by William Lavender.
#
#---------------------------------------------------------------------
#
#  Copyright 1999 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

namespace eval MxOptimize {
	variable logfile_name

	variable monospaced_font
}

proc setup_optimize_database { motor_database scan_database } {

	#
	# Create a new record list.
	#

	set record_list [ Mx::RecordList ::mxrecordlist_#auto ]

	#
	# Enable plotting.
	#

	$record_list set_plot_enable 1

	#
	# Read the motor and scan databases.
	#

	$record_list read_database $motor_database
	$record_list read_database $scan_database

	#
	# Finish initializing MX.
	#

	$record_list finish_database_initialization
	$record_list initialize_hardware

	#
	# Install an event handler to allow Tk events to be processed
	# while MX is scanning.
	#

	Mx::install_user_interrupt_handler

	#
	# Install handler for mx_info_dialog() for things like
	# end of scan dialogs.
	#

	Mx::install_dialog_handler

	return $record_list
}

proc perform_optimization { record_list motor guifield } {

	set logfile_format "%-16s %-16s %f %s"

	set motor_name [ $motor get_field "name" ]

	#
	# For each motor there should be a scan with the same name, but
	# with _scan tacked onto the end of it.
	#

	set scan_name ${motor_name}_scan

	set status [ catch { $record_list get_record $scan_name } scan_record ]

	if { $status != 0 } {
		set t "The scan '$scan_name' does not exist.  "
		append t "I cannot perform the optimization without it."

		tk_dialog .scan_does_not_exist_dialog "Scan Record Missing" \
			$t error 0 "OK"

		return
	}

	#
	# Save the old motor position.
	#

		set old_position [ optimize_get_position $motor ]

		set units [ $motor get_field "units" ]

	#
	# Update the screen display with the most current old position.
	#

		${guifield}_old_position configure -text $old_position

	#
	# Calculate the scan range.
	#
		set step_size [ $scan_record get_field "step_size" ]
		set num [ $scan_record get_field "num_measurements" ]

		set scan_range [ expr ($num - 1) * $step_size ]

	#
	# Change the scan start position so that the scan range is centered
	# on the original position of the motor.
	#

		set scan_start [ expr $old_position - 0.5 * $scan_range ]

		$scan_record set_field "start_position" $scan_start
	#
	# Run the scan.
	#

	${guifield}_status configure -fg yellow \
			-text "Optimization scan in progress"

#
#	set status [ catch { $scan_record perform_scan } ]
#

	set status [ optimize_perform_scan $scan_record ]

	if { $status != 0 } {
		${guifield}_status configure -fg hotpink \
			-text "Moving to original start position."

		log_puts \
"Optimization scan failed.  Moving to old position $old_position $units"

		logfile_puts [ format $logfile_format \
			$motor_name SCAN_FAILED $old_position $status ]

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "scan_failed"

		set t "The scan '$scan_name' failed with status = $status.  "
		append t "Moving '$motor_name' back to its original position."

		tk_dialog .scan_failed_dialog "Scan Failed" \
			$t error 0 "OK"

		return
	}

	#
	# Perform the fit with an external script.
	#

	set program "optimize_peak_fitting"

	set datafile_name [ $scan_record get_field "datafile_name" ]

	set num_lines [ eval exec "wc -l $datafile_name" ]

	scan $num_lines "%d" num_lines

	#

	set command "$program $num_lines $datafile_name"

# 	log_puts "fitting command = '$command'"

	${guifield}_status configure -fg yellow -text "Fitting in progress"

	#
	# Run the script.
	#

	set status [ catch { eval exec $command } result ]

# 	log_puts "status = '$status', result = '$result'"

	log_puts "Fit result = '$result'"

	if { $status != 0 } {
		${guifield}_status configure -fg hotpink \
				-text "Moving to original start position"

		log_puts \
"Fitting program crashed.  Moving to old position $old_position $units"

		logfile_puts [ format $logfile_format \
			$motor_name FIT_CRASHED $old_position $result ]

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "fit_crashed"

		set t "The external fit program crashed with status = $status."
		append t "  Moving '$motor_name' back to its "
		append t "position."

		tk_dialog .fit_crashed_dialog "Fit Crashed" \
			$t error 0 "OK"

		return
	}

	#
	# Convert the result to lower case and replace any D format characters
	# with e.
	#

	set result [ string tolower $result ]

	set length [ string length $result ]

	for { set i 0 } { $i < $length } { incr i } {

		set thischar [ string index $result $i ]

		if { $thischar == "d" } {

			set thischar "e"
		}

		append new_result $thischar
	}

#	log_puts "new_result = '$new_result'"

	#
	# Did the fit work?  If so, move the motor to the new position.
	#

	scan $new_result "%d %g %d %g %g %g %g %g" \
		fit_status euclidean_norm algorithm_info max_position \
		a1 a2 a3 a4

#	log_puts "fit_status = '$fit_status'"
#	log_puts "euclidean_norm = '$euclidean_norm'"
#	log_puts "algorithm_info = '$algorithm_info'"
#	log_puts "max_position = '$max_position'"
#	log_puts "a1 = '$a1'"
#	log_puts "a2 = '$a2'"
#	log_puts "a3 = '$a3'"
#	log_puts "a4 = '$a4'"

	set new_position $a3

	if { $fit_status == 0 } {
		${guifield}_status configure -fg hotpink \
				-text "Moving to original start position"

		log_puts \
"Optimization fit failed.  Moving to old position $old_position $units"

		logfile_puts [ format $logfile_format \
			$motor_name FIT_FAILED $old_position $new_result ]

		$motor move_absolute $old_position

		after 1000 \
			check_final_move_status $motor $guifield "fit_failed"
	} else {
		set gnuplot [ plot_fit $datafile_name $a1 $a2 $a3 $a4 $units ]

		set dialog .optimization_succeeded_dialog

		set dialog_font [ font create -family Times -size 15 ]

		toplevel $dialog
		wm title $dialog "Optimization Succeeded"

		label $dialog.message1 -font $dialog_font \
				-text "The optimization was successful."
		pack $dialog.message1

		set mtext "The new suggested position for '$motor_name' "
		append mtext "is $new_position $units."

		label $dialog.message2 -font $dialog_font -text $mtext
		pack $dialog.message2

		frame $dialog.sep -height 2 -borderwidth 1 -relief sunken
		pack $dialog.sep -expand yes -fill both

		iwidgets::buttonbox $dialog.bb
		pack $dialog.bb -expand yes -fill both

		set logfile_accept_string [ format $logfile_format \
			$motor_name SUCCESS $new_position "" ]

		set logfile_discard_string [ format $logfile_format \
			$motor_name FIT_DISCARDED $old_position "" ]

		$dialog.bb add Accept -text Accept -command [ code \
			optimization_succeeded_accept $dialog $guifield \
			$motor $new_position $units $gnuplot \
			$logfile_accept_string ]

		$dialog.bb add Discard -text Discard -command [ code \
			optimization_succeeded_discard $dialog $guifield \
			$motor $old_position $units $gnuplot \
			$logfile_discard_string ]

		$dialog.bb default Accept
	}
}

proc plot_fit { datafile_name a1 a2 a3 a4 units } {

	set equation \
		"$a1 + $a2 * exp( - 0.5 * (x - $a3)**2 / ($a4**2) )"

	set gnuplot [ open "| gnuplot >& /dev/null" w ]

	puts $gnuplot "set terminal x11"

	puts $gnuplot "set nokey"

	puts $gnuplot "plot '$datafile_name'"

	puts $gnuplot "replot $equation"

	puts $gnuplot \
"set arrow from first $a3, graph 0 to first $a3, graph 1 nohead linetype 3"

	puts $gnuplot \
"set title 'Datafile = $datafile_name     Fitted peak at $a3 $units'"

	puts $gnuplot "replot"

	flush $gnuplot

	return $gnuplot
}

proc optimization_succeeded_accept { dialog guifield motor new_position \
					units gnuplot logfile_message } {

	destroy $dialog

	puts $gnuplot "exit"

	close $gnuplot

	${guifield}_status configure -fg springgreen \
				-text "Moving to new position."

	${guifield}_new_position configure -text $new_position

	log_puts \
"Optimization succeeded.  Moving to optimized position $new_position $units"

	logfile_puts $logfile_message

	$motor move_absolute $new_position

	after 1000 check_final_move_status $motor $guifield "success"
}

proc optimization_succeeded_discard { dialog guifield motor old_position \
					units gnuplot logfile_message } {

	destroy $dialog

	puts $gnuplot "exit"

	close $gnuplot

	${guifield}_status configure -fg purple \
				-text "Moving to original start position."

	${guifield}_new_position configure -text $old_position

	set message "Optimization succeeded, but was discarded.  "
	append message "Moving to original position $old_position $units"

	log_puts $message

	logfile_puts $logfile_message

	$motor move_absolute $old_position

	after 1000 check_final_move_status $motor $guifield "discarded"
}

proc check_final_move_status { motor guifield condition } {

	# See if the motor is still moving.

	set status [ catch { $motor is_busy } result ]

	if { $status == 0 } {

		if { $result != 0 } {

			after 1000 \
			  check_final_move_status $motor $guifield $condition

			return
		}
	} else {
		set condition "device_io_error"

		catch { $motor soft_abort }
	}

	if { $condition == "scan_failed" } {

		${guifield}_status configure -fg red \
				-text "Optimization scan failed."

	} elseif { $condition == "fit_crashed" } {

		${guifield}_status configure -fg red \
				-text "Fitting program crashed."

	} elseif { $condition == "fit_failed" } {

		${guifield}_status configure -fg red \
				-text "Optimization fit failed."

	} elseif { $condition == "device_io_error" } {

		${guifield}_status configure -fg red \
				-text "Motor device I/O error."

	} elseif { $condition == "success" } {

		${guifield}_status configure -fg limegreen \
				-text "Optimization succeeded."

	} elseif { $condition == "discarded" } {

		${guifield}_status configure -fg maroon \
				-text "Optimization discarded."

	} else { 
		${guifield}_status configure -fg red \
	-text "check_final_move_status invoked with condition '$condition'"
	}
}

proc optimize_get_position { motor } {

	set position [ $motor get_position ]

	set position [ format {%.8g} $position ]

	return $position
}

###########################################################################
#                                                                         #
# optimize_perform_scan is used to run a scan and modify the top level gui#
# display as necessary.                                                   #
#                                                                         #
###########################################################################

proc optimize_perform_scan { scan_record } {

	set scan_name [ $scan_record cget -name ]

	set window1 .scan_control_window
	set window2 .scan_control_window_sep

	frame $window1
	pack $window1 -before .message_log

	frame $window2 -height 2 -borderwidth 1 -relief sunken
	pack $window2 -pady 5 -expand yes -fill x -before .message_log

	set scan_frame $window1.frame

	label $window1.scan_label -text "Current scan: "

	label $window1.scan_name -text $scan_name \
				-font $MxOptimize::monospaced_font

	pack $window1.scan_label $window1.scan_name -side left

	frame $window1.sep 
	pack $window1.sep -side left -expand yes

	label $window1.sep.junk -width 5
	pack $window1.sep.junk

	button $window1.abort_scan -background red -activebackground red \
		-text "Abort Scan" -command [ code optimize_abort_scan ]

	pack $window1.abort_scan -side right

	update

	set status [ catch { $scan_record perform_scan } ]

	update

	pack forget $window1
	destroy $window1

	pack forget $window2
	destroy $window2

	update

	return $status
}

proc optimize_abort_scan {} {
	log_puts "Scan abort requested."

	set Mx::user_interrupt 1
}

proc create_message_log_window {} {

	set window .message_log

        #
        #  Create the window.
        #

        iwidgets::scrolledtext $window -labeltext "Message Log"

        $window configure -visibleitems 40x10
        $window configure -hscrollmode dynamic
        $window configure -vscrollmode dynamic

        pack $window -expand yes -fill both

        #
        #  Install handlers to redirect the output of mx_info(), mx_error()
        #  and MX_DEBUG() to this window.
        #

        Mx::install_output_handlers $window

	log_puts "*** Optimize GUI Initializing ***"
}

proc log_puts { message } {
	set textwindow [ .message_log childsite ]

	$textwindow configure -state normal

	.message_log insert end "$message\n"

	$textwindow configure -state disabled

	.message_log yview moveto 1.0
}

proc logfile_puts { message } {

	# Log file messages are only recorded if $MxOptimize::logfile_name
	# has a length greater than zero.

	if { $MxOptimize::logfile_name == "" } {
		return
	}

	#
	# Construct a timestamp to write to the log file.
	#

	set current_time [ clock seconds ]

	set timestamp [ clock format $current_time -format \
		"%d/%m/%Y %H:%M:%S " ]

	#
	# Finish by writing the message to the log file.
	#

	set logfile [ open $MxOptimize::logfile_name "a" ]

	puts $logfile "$timestamp $message"

	close $logfile

	return
}

